/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.base;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLContext;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.Loadable;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.FaceList;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.LineFace;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.PointFace;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.QuadFace;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.TriangleFace;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.LinkedList;

/**
 * Used to store vertices and faces to draw a 3D object.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 * @see Face
 * @see DrawingType
 */
public abstract class Primitive extends Loadable {

    /**
     * The list of vertices.
     */
    private final VertexList vertices;
    
    /**
     * The list of POINT drawing mode faces.
     * @see Face
     * @see DrawingType
     */
    private final FaceList pointFaces;
    
    /**
     * The list of LINE drawing mode faces.
     * @see Face
     * @see DrawingType
     */
    private final FaceList lineFaces;
    
    /**
     * The list of triangles drawing mode faces.
     * @see Face
     * @see DrawingType
     */
    private final FaceList triangleFaces;
    
    /**
     * The list of quad (square) drawing mode faces.
     * @see Face
     * @see DrawingType
     */
    private final FaceList quadFaces;

    /**
     * Vertex array object.
     */
    private final IntBuffer vao;
    
    /**
     * Vertex position buffer object.
     */
    private final IntBuffer vpbo;
    
    /**
     * Vertex normal buffer object.
     */
    private final IntBuffer vnbo;

    /**
     * Create an empty primitive object.
     */
    public Primitive() {
        this.vertices = new VertexList();
        this.pointFaces = new FaceList();
        this.lineFaces = new FaceList();
        this.triangleFaces = new FaceList();
        this.quadFaces = new FaceList();

        this.vao = Buffers.newDirectIntBuffer(1);
        this.vpbo = Buffers.newDirectIntBuffer(1);
        this.vnbo = Buffers.newDirectIntBuffer(1);
    }

    /**
     * Add a vertex (use-it in the generateVertices function)
     * @param vertex 
     */
    public final void addVertex(final Vertex vertex) {
        vertices.addVertex(vertex);
    }

    /**
     * Add a point face (use-it in the generatePointFace function)
     * @param face 
     */
    public final void addPointFace(final PointFace face) {
        pointFaces.addFace(face);
    }

    /**
     * Add a line face (use-it in the generateLineFace function)
     * @param face 
     */
    public final void addLineFace(final LineFace face) {
        lineFaces.addFace(face);
    }

    /**
     * Add a triangle face (use-it in the generateTriangleFace function)
     * @param face 
     */
    public final void addTriangleFace(final TriangleFace face) {
        triangleFaces.addFace(face);
    }

    /**
     * Add a quad (square) face (use-it in the generateQuadFace function)
     * @param face 
     */
    public final void addQuadFace(final QuadFace face) {
        quadFaces.addFace(face);
    }

    /**
     * Get the number of vertex.
     * Can be called meny times.
     * @return 
     */
    public final int getVertexCount() {
        if (vertices.isEmpty()) {
            generateVertices();
        }

        return vertices.getLength();
    }

    /**
     * Get the vertices buffer.
     * Can be called meny times.
     * @return 
     */
    public final FloatBuffer getVertexPositionBuffer() {
        if (vertices.isEmpty()) {
            generateVertices();
        }

        return Buffers.newDirectFloatBuffer(vertices.getPositions());
    }

    /**
     * Implements-it to generate each vertex.
     */
    protected abstract void generateVertices();

    /**
     * Get if this primitive has line faces or not.
     * @return 
     */
    public final boolean hasPointFaces() {
        return !pointFaces.isEmpty();
    }
    
    /**
     * Get the point index buffer.
     * Can be called meny times.
     * @return 
     */
    public final IntBuffer getPointIndexBuffer() {
        if (!hasPointFaces()) {
            generatePointFaces();
        }

        return Buffers.newDirectIntBuffer(pointFaces.getIndices());
    }

    /**
     * Implement-it to generate point faces.
     */
    protected abstract void generatePointFaces();

    /**
     * Get if this primitive has line faces or not.
     * @return 
     */
    public final boolean hasLineFaces() {
        return !lineFaces.isEmpty();
    }
    
    /**
     * Get the line index buffer.
     * Can be called meny times.
     * @return 
     */
    public final IntBuffer getLineIndexBuffer() {
        if (!hasLineFaces()) {
            generateLineFaces();
        }

        return Buffers.newDirectIntBuffer(lineFaces.getIndices());
    }

    /**
     * Implements it to generate line faces.
     */
    protected abstract void generateLineFaces();

    /**
     * Get if this primitive has triangle faces or not.
     * @return 
     */
    public final boolean hasTriangleFaces() {
        return !triangleFaces.isEmpty();
    }
    
    /**
     * Get the triangle index buffer.
     * Can be called many times.
     * @return 
     */
    public final IntBuffer getTriangleIndexBuffer() {
        if (!hasTriangleFaces()) {
            generateTriangleFaces();
        }

        return Buffers.newDirectIntBuffer(triangleFaces.getIndices());
    }

    /**
     * Implement it to generate triangle faces.
     */
    protected abstract void generateTriangleFaces();

    /**
     * Get if this primitive has quad faces or not.
     * @return 
     */
    public final boolean hasQuadFaces() {
        return !quadFaces.isEmpty();
    }
    
    /**
     * Get the quad index buffer.
     * Can be called many times.
     * @return 
     */
    public final IntBuffer getQuadIndexBuffer() {
        if (!hasQuadFaces()) {
            generateQuadFaces();
        }

        return Buffers.newDirectIntBuffer(quadFaces.getIndices());
    }

    /**
     * Implement it to generate quad faces.
     */
    protected abstract void generateQuadFaces();

    @Override
    protected final void onLoad() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        if (getVertexCount() > 0) {
            /* Create vertex arrays */
            gl.glGenVertexArrays(vao.limit(), vao);
            gl.glBindVertexArray(vao.get(0));
        }
    }
    
    private void loadPositions() {
        bind();
        
        if (vpbo.get(0) == 0) {
            final GL3bc gl = GLContext.getCurrentGL().getGL3bc();

            /* Create vertex position buffers */
            gl.glGenBuffers(vpbo.limit(), vpbo);
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vpbo.get(0));
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
                    getVertexPositionBuffer().limit() * Float.BYTES,
                    getVertexPositionBuffer(), GL3bc.GL_STATIC_DRAW);
        }
    }
    
    private void loadNormals() {
        bind();
        
        if (vnbo.get(0) == 0) {
            final GL3bc gl = GLContext.getCurrentGL().getGL3bc();

            /* Create vertex normal buffers */
            gl.glGenBuffers(vnbo.limit(), vnbo);
        }
    }

    @Override
    protected final void onBind() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();

        if (getVertexCount() > 0) {
            /* Bind vertex arrays */
            gl.glBindVertexArray(vao.get(0));
        }
    }
    
    public final void bindPositions(final int attribID) {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        loadPositions();

        /* Enable vertex position attribute */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vpbo.get(0));
        gl.glEnableVertexAttribArray(attribID);
        gl.glVertexAttribPointer(0, 3, GL3bc.GL_FLOAT, false, 0, 0);
    }
    
    public final void bindPointsNormals(final int attribID) {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        if (isLoaded()) {
            loadNormals();
            
            if (!hasPointFaces()) {
                generatePointFaces();
            }
        
            final FloatBuffer buffer = 
                    Buffers.newDirectFloatBuffer(pointFaces.getNormals());

            /* Enable vertex normal attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vnbo.get(0));
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
            buffer.limit() * Float.BYTES,
            buffer, GL3bc.GL_STATIC_DRAW);
            gl.glEnableVertexAttribArray(attribID);
            gl.glVertexAttribPointer(attribID, 3, GL3bc.GL_FLOAT, false, 0, 0);
        }
    }
    
    public final void bindLinesNormals(final int attribID) {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        if (isLoaded()) {
            loadNormals();

            if (!hasLineFaces()) {
                generateLineFaces();
            }
        
            final FloatBuffer buffer = 
                    Buffers.newDirectFloatBuffer(lineFaces.getNormals());

            /* Enable vertex normal attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vnbo.get(0));
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
            buffer.limit() * Float.BYTES,
            buffer, GL3bc.GL_STATIC_DRAW);
            gl.glEnableVertexAttribArray(attribID);
            gl.glVertexAttribPointer(attribID, 3, GL3bc.GL_FLOAT, false, 0, 0);
        }
    }
    
    public final void bindTrianglesNormals(final int attribID) {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        if (isLoaded()) {
            loadNormals();

            if (!hasTriangleFaces()) {
                generateLineFaces();
            }
        
            final FloatBuffer buffer = 
                    Buffers.newDirectFloatBuffer(triangleFaces.getNormals());

            /* Enable vertex normal attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vnbo.get(0));
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
            buffer.limit() * Float.BYTES,
            buffer, GL3bc.GL_STATIC_DRAW);
            gl.glEnableVertexAttribArray(attribID);
            gl.glVertexAttribPointer(attribID, 3, GL3bc.GL_FLOAT, false, 0, 0);
        }
    }
    
    public final void bindQuadsNormals(final int attribID) {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        if (isLoaded()) {
            loadNormals();

            if (!hasQuadFaces()) {
                generateQuadFaces();
            }
        
            final FloatBuffer buffer = 
                    Buffers.newDirectFloatBuffer(quadFaces.getNormals());

            /* Enable vertex normal attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vnbo.get(0));
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
            buffer.limit() * Float.BYTES,
            buffer, GL3bc.GL_STATIC_DRAW);
            gl.glEnableVertexAttribArray(attribID);
            gl.glVertexAttribPointer(attribID, 3, GL3bc.GL_FLOAT, false, 0, 0);
        }
    }
    
    @Override
    protected void onUnbind() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        gl.glBindVertexArray(0);
    }

    @Override
    protected final void onUnload() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();

        if (getVertexCount() > 0) {
            /* Clean up */
            gl.glDeleteBuffers(vpbo.capacity(), vpbo);
            gl.glDeleteVertexArrays(vao.capacity(), vao);
        }
    }

}
