/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.loadables.programs;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLContext;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.base.DrawingType;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.base.Mesh;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.extended.CubePrimitive;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.shaders.VoxelShader;
import java.awt.Color;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import org.joml.Matrix4f;

/**
 *
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class VoxelProgram extends Program {
    
        
    /* GLSL uniforms area */
    
    /**
     * Projection matrix uniform value.
     */
    private Matrix4f proj;
    
    /**
     * Model matrix uniform value.
     */
    private Matrix4f model;
    
    /**
     * Scale matrix uniform value.
     */
    private Matrix4f scale;
    
    /**
     * Voxel space width uniform value.
     */
    private int width;
    
    /**
     * Voxel space height uniform value.
     */
    private int height;
    
    /**
     * Voxel space depth uniform value.
     */
    private int depth;
    
    /* GLSL attributes area */
    
    /**
     * Voxels offsets attribute value.
     */
    private float[] offsets;
    
    /**
     * Voxel color values.
     */
    private Color[] colors;
    
    /* OpenGL uniform locations area */
    private int projLoc;
    private int modelLoc;
    private int scaleLoc;
    private int widthLoc;
    private int heightLoc;
    private int depthLoc;
    
    /**
     * Additionnal vertex buffer object (used for the attributes).
     */
    private IntBuffer vbo;

    public VoxelProgram() {
        super();
        
        this.shader = VoxelShader.getINSTANCE();
        this.mesh = new Mesh(CubePrimitive.getINSTANCE(), DrawingType.QUADS);
        
        this.proj = new Matrix4f().identity();
        this.model = new Matrix4f().identity();
        this.scale = new Matrix4f().identity();
        this.width = 0;
        this.height = 0;
        this.depth = 0;
        this.offsets = new float[0];
        this.colors = new Color[0];
        
        this.vbo = Buffers.newDirectIntBuffer(2);
    }
    
    public final void setProj(final Matrix4f proj) {
        this.proj = proj;
    }
    
    public final void setModel(final Matrix4f model) {
        this.model = model;
    }
    
    public final void setScale(final Matrix4f scale) {
        this.scale = scale;
    }

    public final void setWidth(final int width) {
        this.width = width;
    }

    public final void setHeight(final int height) {
        this.height = height;
    }

    public final void setDepth(final int depth) {
        this.depth = depth;
    }

    public final void setOffsets(final float... offsets) {
        this.offsets = offsets;
    }

    public final void setColors(final Color... colors) {
        this.colors = colors;
    }

    @Override
    protected void onLoad() {
        super.onLoad();
        
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        /* Get the uniform locations */
        projLoc = gl.glGetUniformLocation(shader.getProgram(), "proj");
        modelLoc = gl.glGetUniformLocation(shader.getProgram(), "model");
        scaleLoc = gl.glGetUniformLocation(shader.getProgram(), "scale");
        widthLoc = gl.glGetUniformLocation(shader.getProgram(), "width");
        heightLoc = gl.glGetUniformLocation(shader.getProgram(), "height");
        depthLoc = gl.glGetUniformLocation(shader.getProgram(), "depth");
        
        /* Create the instancing attributes */
        mesh.bind();
        gl.glGenBuffers(2, vbo);
        
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
        gl.glVertexAttribPointer(1, 3, GL3bc.GL_FLOAT, false, 0, 0);
        gl.glEnableVertexAttribArray(1);
        
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
        gl.glVertexAttribPointer(2, 3, GL3bc.GL_FLOAT, false, 0, 0);
        gl.glEnableVertexAttribArray(2);
    }

    @Override
    protected void onBind() {
        super.onBind();
        
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        /* Send uniform */
        gl.glUniformMatrix4fv(projLoc, 1, false, proj.get(new float[16]), 0);
        gl.glUniformMatrix4fv(modelLoc, 1, false, model.get(new float[16]), 0);
        gl.glUniformMatrix4fv(scaleLoc, 1, false, scale.get(new float[16]), 0);
        gl.glUniform1i(widthLoc, width);
        gl.glUniform1i(heightLoc, height);
        gl.glUniform1i(depthLoc, depth);
        
        /* Send vertex attributes */
        mesh.bindPositions(0);
        
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
        final FloatBuffer offsetBuf = Buffers.newDirectFloatBuffer(offsets);
        gl.glBufferData(GL3bc.GL_ARRAY_BUFFER, 
                Float.BYTES * offsetBuf.limit(), offsetBuf, GL3bc.GL_STATIC_DRAW);
        gl.glVertexAttribDivisor(1, 1);
        
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
        final float[] colorf = new float[colors.length * 3];
        for (int i = 0; i < colors.length; i++) {
            colorf[i * 3 + 0] = (float) colors[i].getRed() / 255;
            colorf[i * 3 + 1] = (float) colors[i].getGreen() / 255;
            colorf[i * 3 + 2] = (float) colors[i].getBlue() / 255;
        }
        final FloatBuffer colorsBuf = Buffers.newDirectFloatBuffer(colorf);
        gl.glBufferData(GL3bc.GL_ARRAY_BUFFER, 
                Float.BYTES * colorsBuf.limit(), colorsBuf, GL3bc.GL_STATIC_DRAW);
        gl.glVertexAttribDivisor(2, 1);
        
        /* Draw the number of instance */
        mesh.setInstances(colors.length);
        mesh.draw();
    }

    @Override
    protected void onUnload() {
        super.onUnload();
        
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        /* Clean up */
        gl.glDeleteBuffers(vbo.limit(), vbo);
    }
    
}
