/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.loadables.shaders;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLContext;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.Loadable;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * GLSL shader class to load, use and unload OpenGL programs.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class Shader extends Loadable {
    
    /**
     * Current binded program (used for optimization).
     */
    private static int bindedProgram = 0;
    
    /**
     * The program it attached with this shader.
     */
    private int program;
    
    /**
     * The vertex shader id attached with this object.
     */
    private int vertexShader;
    
    /**
     * The fragment shader id attached with this object.
     */
    private int fragmentShader;
    
    /**
     * The vertex GLSL source code.
     */
    private String[] vertexSources;
    
    /**
     * The fragment GLSL source code.
     */
    private String[] fragmentSources;

    /**
     * Create an empty shader.
     */
    public Shader() {
        this.program = 0;
        this.vertexShader = 0;
        this.fragmentShader = 0;
        this.vertexSources = null;
        this.fragmentSources = null;
    }

    /**
     * Get the opengl program id.
     * @return 
     */
    public int getProgram() {
        return program;
    }

    /**
     * Change the vertex sources.
     * @param sources 
     */
    public final void setVertexSources(final String[] sources) {
        this.vertexSources = sources;
    }

    /**
     * Change the fragment sources.
     * @param fragmentSources 
     */
    public final void setFragmentSources(final String[] fragmentSources) {
        this.fragmentSources = fragmentSources;
    }
    
    /**
     * Import the sources of a GLSL file.
     * @param path
     * @return 
     */
    public final String[] getGLSLSources(final String path) {
        String sources = "";

        try {
            InputStream input = getClass().getResourceAsStream(path);
            InputStreamReader inputReader = new InputStreamReader(input);
            BufferedReader reader = new BufferedReader(inputReader);

            String st;
            while ((st = reader.readLine()) != null) {
                sources += st + "\n";
            }
        } catch (IOException ex) {
            Logger.getLogger(getClass().getName()).log(Level.ERROR, null, ex);
        }
        
        return sources.split("(?=\n)");
    }
    
    /**
     * Initialize an opengl shader with the specified type and sources.
     * @param type The type of shader to init.
     * @param sources The GLSL source code.
     * @return The opengl shader id.
     */
    private int initShader(final int type, final String[] sources) {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        /* Create, attach source to, and compile shader */
        final int shader = gl.glCreateShader(type);
        gl.glShaderSource(shader, sources.length, sources, null, 0);
        gl.glCompileShader(shader);
        
        /* Get the compilation status */
        final IntBuffer compilationStatus = Buffers.newDirectIntBuffer(1);
        gl.glGetShaderiv(shader, GL3bc.GL_COMPILE_STATUS, compilationStatus);

        /* If the compilation has failed, print error logs */
        if (compilationStatus.get() != GL3bc.GL_TRUE) {
            /* Get the info log length */
            final IntBuffer logLength = Buffers.newDirectIntBuffer(1);
            gl.glGetShaderiv(shader, GL3bc.GL_INFO_LOG_LENGTH, logLength);

            /* Get the log, and print-it */
            ByteBuffer infoLog = Buffers.newDirectByteBuffer(logLength.get(0));
            gl.glGetShaderInfoLog(shader, logLength.get(0), logLength, infoLog);
            byte[] b = new byte[logLength.get()];
            infoLog.get(b);
            System.err.println("Error compiling shader n�" + shader 
                    + " -> " + new String(b));
        }
        
        return shader;
    }
    
    @Override
    protected final void onLoad() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
                
        if (vertexSources != null && fragmentSources != null) {
            vertexShader = initShader(GL3bc.GL_VERTEX_SHADER, vertexSources);
            fragmentShader = initShader(GL3bc.GL_FRAGMENT_SHADER, fragmentSources);
            
            program = gl.glCreateProgram();
            gl.glAttachShader(program, vertexShader);
            gl.glAttachShader(program, fragmentShader);
            gl.glLinkProgram(program);
            
            /* Get the linking status */
            final IntBuffer linkStatus = Buffers.newDirectIntBuffer(1);
            gl.glGetProgramiv(program, GL3bc.GL_LINK_STATUS, linkStatus);
            
            if (linkStatus.get() != GL3bc.GL_TRUE) {
                /* Get the info log length */
                final IntBuffer logLength = Buffers.newDirectIntBuffer(1);
                gl.glGetProgramiv(program, GL3bc.GL_INFO_LOG_LENGTH, logLength);

                /* Get the log, and print-it */
                final ByteBuffer infoLog = Buffers.newDirectByteBuffer(logLength.get(0));
                gl.glGetProgramInfoLog(program, logLength.get(0), logLength, infoLog);
                final byte[] b = new byte[logLength.get()];
                infoLog.get(b);
                System.err.println("Error linking program n�" + program 
                        + " -> " + new String(b));
            }
        }
    }

    @Override
    protected void onBind() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        /* Bind the attached program if that is not already binded */
        if (bindedProgram != program) {
            gl.glUseProgram(program);
            bindedProgram = program;
        }
    }

    @Override
    protected void onUnbind() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        gl.glUseProgram(0);
    }

    @Override
    protected final void onUnload() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        /* Clean up shaders */
        if (vertexShader > 0) {
            gl.glDeleteShader(vertexShader);
        }
        
        if (fragmentShader > 0) {
            gl.glDeleteShader(fragmentShader);
        }
         /* Clean up program */
        if (program > 0) {
            gl.glDeleteProgram(program);
        }
    }

}
