/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.voxels;

import fr.amap.lidar.amapvox.visualization.controllers.loadables.programs.BorderProgram;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.programs.VoxelProgram;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.programs.SimpleProgram;
import com.jogamp.newt.event.MouseEvent;
import fr.amap.commons.util.ColorGradient;
import fr.amap.lidar.amapvox.visualization.controllers.inputs.Input;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.base.DrawingType;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.base.Mesh;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.extended.AxisPrimitive;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.extended.PointPrimitive;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.programs.SkyboxProgram;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.textures.TextureCubemap;
import java.awt.Color;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import org.joml.Matrix4f;
import org.joml.Quaternionf;
import org.joml.Vector4f;

/**
 * Voxel space 3D class. This is the main class of the 3D view. It can be
 * rotated to display each voxel has a group of cube.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class Space {

    /**
     * The attached program.
     */
    private final VoxelProgram pointProgram;
    
    /**
     * The attached program.
     */
    private final VoxelProgram cubeProgram;
    
    /**
     * The attached axis program.
     */
    private final SimpleProgram axisProgram;

    /**
     * The attached border program.
     */
    private final BorderProgram borderProgram;
    
    /**
     * The attached skybox program.
     */
    private final SkyboxProgram skyboxProgram;

    /**
     * The attached voxel space file.
     */
    private final VoxelSpaceFile file;

    /**
     * Filters list (used to hide some voxels).
     */
    private final ConcurrentHashMap<String, Filter> filters;

    /**
     * Mouse rotation (used to perform a linear interpolation).
     */
    private Quaternionf mouseRotation;

    /**
     * The rotation of this object.
     */
    private Quaternionf rotation;
    
    /**
     * The size of each voxel drawables.
     */
    private float voxelSize;

    /**
     * The current colorbar.
     */
    private Color[] colorBar;
    
    /**
     * The minimum value of the renderer property.
     */
    private float colorMin;
    
    /**
     * The maximum value of the renderer property.
     */
    private float colorMax;
    
    private float fov;

    /**
     * Create a new voxel space with the specified voxel space file.
     * @param file
     */
    public Space(final VoxelSpaceFile file) throws IOException {
        this.pointProgram = new VoxelProgram();
        pointProgram.setMesh(new Mesh(PointPrimitive.getINSTANCE(), DrawingType.POINTS));
        
        this.cubeProgram = new VoxelProgram();
        
        this.borderProgram = new BorderProgram();
        
        this.axisProgram = new SimpleProgram();
        axisProgram.setMesh(new Mesh(AxisPrimitive.getINSTANCE(), DrawingType.LINES));
        
        this.skyboxProgram = new SkyboxProgram();
        skyboxProgram.setCubemap(
        new TextureCubemap(512, 512, 
                "/textures/skybox/right.png", "/textures/skybox/left.png", 
                "/textures/skybox/top.png", "/textures/skybox/bottom.png", 
                "/textures/skybox/back.png", "/textures/skybox/front.png"));
        
        this.file = file;
        this.mouseRotation = new Quaternionf();
        this.rotation = new Quaternionf();
        this.filters = new ConcurrentHashMap();
        this.colorBar = ColorGradient.GRADIENT_GREEN_YELLOW_ORANGE_RED;
        this.voxelSize = 0.5f;
        this.colorMin = 0f;
        this.colorMax = 1f;
        this.fov = 60;

        setProperty(file.getNames()[0]);
        pointProgram.setWidth(file.getWidth());
        pointProgram.setHeight(file.getHeight());
        pointProgram.setDepth(file.getDepth());
        
        cubeProgram.setWidth(file.getWidth());
        cubeProgram.setHeight(file.getHeight());
        cubeProgram.setDepth(file.getDepth());

        borderProgram.setWidth(file.getWidth());
        borderProgram.setHeight(file.getHeight());
        borderProgram.setDepth(file.getDepth());
        
        axisProgram.setMainColor(Color.BLACK);
    }

    /**
     * Add or remplace a property filter
     * @param property
     * @param filter
     */
    public final void setFilter(final String property, final Filter filter) {
        filters.put(property, filter);
    }

    /**
     * Get the filter of the specifief property name (if exist).
     * @param property
     * @return
     */
    public final Filter getFilter(final String property) {
        return filters.get(property);
    }

    /**
     * Get the voxel space file.
     * @return
     */
    public VoxelSpaceFile getFile() {
        return file;
    }

    /**
     * Remove each current filter.
     */
    public final void resetFilters() {
        filters.clear();
    }

    /**
     * Set the current displayed property.
     * @param property
     */
    public final void setProperty(final String property) {
        if (property != null) {
            /* Create two lists for the values 
            and the locations of the voxels */
            final List<Float> offsetList = new ArrayList();
            final List<Color> colorList = new ArrayList();

            /* Get the min and max value of the new property */
            final float maxValue = file.getMaxPropertyValue(property);
            final float minValue = file.getMinPropertyValue(property);

            /* Get each voxel's value of the new property */
            for (int i = 0; i < file.getWidth() * file.getHeight() * file.getDepth(); i++) {
                final float value = Float.valueOf(file.getProperty(property, i));

                /* Perform a hiding of voxels */
                boolean hiden = false;

                final Enumeration<String> keys = filters.keys();

                /* for each filters */
                while (keys.hasMoreElements()) {
                    final String key = keys.nextElement();

                    /* Get the filter property value */
                    final float v = Float.valueOf(file.getProperty(key, i));

                    /* Hide the filtered voxels */
                    if (filters.get(key).isHiden(v)) {
                        hiden = true;
                    }
                }

                /* Add the voxel property to the list (only if not hiden) */
                if (!Float.isNaN(value)) {
                    if (!hiden) {
                        /* Add offset values by inverting k and j */
                        offsetList.add(Float.valueOf(file.getProperty("i", i))
                                * 2 - file.getWidth() + 0.1f);
                        offsetList.add(Float.valueOf(file.getProperty("k", i))
                                * 2 - file.getDepth() - 0.5f);
                        offsetList.add(Float.valueOf(file.getProperty("j", i))
                                * 2 - file.getHeight() + 0.25f);

                        /* Perform a normalization of the value */
                        final float min = minValue + (maxValue - minValue) * colorMin;
                        final float max = minValue + (maxValue - minValue) * colorMax;
                        float normalizedValue = (value - min) / (max - min);
                        
                        if (normalizedValue < 0f) {
                            normalizedValue = 0f;
                        } else if (normalizedValue > 1f) {
                            normalizedValue = 1f;
                        }
                        
                        /* Add color value from the colorbar */
                        final int colorID
                                = Math.round(normalizedValue * (colorBar.length - 1));
                        colorList.add(colorBar[colorID]);
                    }
                }
            }

            /* Transform the two list by two arrays */
            final float[] offsets = new float[offsetList.size()];
            final Color[] colors = new Color[colorList.size()];

            for (int i = 0; i < offsets.length; i++) {
                offsets[i] = offsetList.get(i);
            }

            for (int i = 0; i < colors.length; i++) {
                colors[i] = colorList.get(i);
            }

            /* Send to the program */
            pointProgram.setOffsets(offsets);
            cubeProgram.setOffsets(offsets);
            pointProgram.setColors(colors);
            cubeProgram.setColors(colors);
        }
    }

    public final void setDrawingMode(final float mode) {
        voxelSize = mode;
    }

    /**
     * Move this object to see the specified view.
     * @param view
     */
    public final void setView(final ViewType view) {
        switch (view) {
            case Z_NEG:
                mouseRotation.setAngleAxis(0, 0, 0, 0);
                break;
            case Z_POS:
                mouseRotation.setAngleAxis((float) Math.toRadians(180), 1, 0, 0);
                break;
            case Y_NEG:
                mouseRotation.setAngleAxis((float) Math.toRadians(90), 1, 0, 0);
                break;
            case Y_POS:
                mouseRotation.setAngleAxis((float) Math.toRadians(-90), 1, 0, 0);
                break;
            case X_NEG:
                mouseRotation.setAngleAxis((float) Math.toRadians(90), 0, 0, 1);
                break;
            case X_POS:
                mouseRotation.setAngleAxis((float) Math.toRadians(-90), 0, 0, 1);
                break;
        }
    }

    /**
     * Change the current colorBar (or gradient).
     * @param colorBar
     */
    public final void setColorBar(final Color[] colorBar) {
        this.colorBar = colorBar;
    }
    
    public final void setColorMin(final float min) {
        this.colorMin = min;
    }

    public final void setColorMax(final float max) {
        this.colorMax = max;
    }

    /**
     * Display each voxels.
     * @param delta Delta time.
     */
    public final void display(final float delta) {
        /* Perform rotation */
        final float mouseX = (float) Input.getMouse().getY() * 0.0004f * delta;
        final float mouseY = (float) Input.getMouse().getX() * 0.0004f * delta;

        final float mouseXLength = mouseX > 0f ? mouseX : -mouseX;
        final float mouseYLength = mouseY > 0f ? mouseY : -mouseY;
        
        if (Input.getButtonPress(MouseEvent.BUTTON1)) {
            mouseRotation.rotateLocalX(mouseX);
            mouseRotation.rotateLocalY(mouseY);
        } else if (Input.getButtonPress(MouseEvent.BUTTON3)) {
            if (mouseXLength > mouseYLength) {
                mouseRotation.rotateLocalX(mouseX);
            } else {
                mouseRotation.rotateLocalY(mouseY);
            }
            
            rotation.set(mouseRotation);
        }
        
        rotation.nlerp(mouseRotation, 0.1f);
        
        /* Perform zoom */
        if (Input.isScrolled()) {
            fov -= Input.getMouse().getScroll();
            
            if (fov > 100) {
                fov = 100;
            } else if (fov < 20) {
                fov = 20;
            }
        }

        /* Draw the skybox */
        skyboxProgram.setModel(new Matrix4f().rotate(rotation));
        skyboxProgram.bind();
        
        /* Draw the border (bounding box) */
        Matrix4f model = rotation.get(new Matrix4f());
        borderProgram.setModel(model.scale(
                file.getWidth() * 1.001f, file.getDepth() * 1.001f, file.getHeight() * 1.001f));
        borderProgram.bind();
        
        /* Check if the draw fast is enabled */
        boolean drawFast = false;
        
        if (rotation.angle() > mouseRotation.angle() + 0.001f 
                || rotation.angle() < mouseRotation.angle() - 0.001f) {
            drawFast = true;
        }

        /* Draw each voxels  */
        if (voxelSize > 0 && !drawFast) {
            /* Draw cube */
            cubeProgram.setModel(new Matrix4f().identity().rotate(rotation));
            cubeProgram.setScale(new Matrix4f().identity().scale(voxelSize * 0.9f + 0.1f));
            cubeProgram.bind();
        } else {
            /* Draw point */
            pointProgram.setModel(new Matrix4f().identity().rotate(rotation));
            pointProgram.setScale(new Matrix4f().identity());
            pointProgram.bind();
        }
        
        /* Draw the origin axis */
        axisProgram.setModel(rotation.get(new Matrix4f()));
        axisProgram.bind();
    }

    /**
     * Rescale the projection matrix.
     * @param width
     * @param height
     */
    public final void rescale(final int width, final int height) {        
        final Matrix4f proj = new Matrix4f().setPerspective(
                (float) Math.toRadians(fov), (float) width / height, 0.01f, 1000f);

        borderProgram.setProj(proj);
        pointProgram.setProj(proj);
        cubeProgram.setProj(proj);
        skyboxProgram.setProj(proj);
        
        
        final Matrix4f proj60FOV = new Matrix4f().setPerspective(
                (float) Math.toRadians(60), (float) width / height, 0.01f, 50f);
        final Vector4f translation = new Vector4f(1.3f, -1.35f, 0, 0).mul(proj60FOV);
        axisProgram.setView(new Matrix4f().translate(10 / translation.x, 10 / translation.y, 0 -10));
        axisProgram.setProj(proj60FOV);
    }

}
